@@ -27,6 +27,7 @@
 #include "context-jsapi.h"
 #include "importer.h"
 #include "jsapi-util.h"
+#include "profiler.h"
 
 #include <util/log.h>
 #include <util/error.h>
@@ -56,6 +57,8 @@ struct _GjsContext {
     JSContext *context;
     JSObject *global;
 
+    GjsProfiler *profiler;
+
     char **search_path;
 
     unsigned int we_own_runtime : 1;
@@ -83,6 +86,11 @@ enum {
     PROP_IS_LOAD_CONTEXT
 };
 
+
+static GStaticMutex contexts_lock = G_STATIC_MUTEX_INIT;
+static GList *all_contexts = NULL;
+
+
 static JSBool
 gjs_log(JSContext *context,
         JSObject  *obj,
@@ -164,6 +172,96 @@ gjs_log_error(JSContext *context,
     return JS_TRUE;
 }
 
+static JSBool
+gjs_print_parse_args(JSContext *context,
+                     uintN      argc,
+                     jsval     *argv,
+                     char     **buffer)
+{
+    GString *str;
+    gchar *s;
+    guint n;
+
+    str = g_string_new("");
+    JS_EnterLocalRootScope(context);
+    for (n = 0; n < argc; ++n) {
+        JSExceptionState *exc_state;
+        JSString *jstr;
+
+        /* JS_ValueToString might throw, in which we will only
+         * log that the value could be converted to string */
+        exc_state = JS_SaveExceptionState(context);
+
+        jstr = JS_ValueToString(context, argv[n]);
+        if (jstr != NULL)
+            argv[n] = STRING_TO_JSVAL(jstr); // GC root
+
+        JS_RestoreExceptionState(context, exc_state);
+
+        if (jstr != NULL) {
+            if (!gjs_string_to_utf8(context, STRING_TO_JSVAL(jstr), &s)) {
+                JS_LeaveLocalRootScope(context);
+                g_string_free(str, TRUE);
+                return JS_FALSE;
+            }
+
+            g_string_append(str, s);
+            g_free(s);
+            if (n < (argc-1))
+                g_string_append_c(str, ' ');
+        } else {
+            JS_LeaveLocalRootScope(context);
+            *buffer = g_string_free(str, TRUE);
+            if (!*buffer)
+                *buffer = g_strdup("<invalid string>");
+            return JS_TRUE;
+        }
+
+    }
+    JS_LeaveLocalRootScope(context);
+    *buffer = g_string_free(str, FALSE);
+
+    return JS_TRUE;
+}
+
+static JSBool
+gjs_print(JSContext *context,
+          JSObject  *obj,
+          uintN      argc,
+          jsval     *argv,
+          jsval     *retval)
+{
+    char *buffer;
+
+    if (!gjs_print_parse_args(context, argc, argv, &buffer)) {
+        return FALSE;
+    }
+
+    g_print("%s\n", buffer);
+    g_free(buffer);
+
+    return JS_TRUE;
+}
+
+static JSBool
+gjs_printerr(JSContext *context,
+             JSObject  *obj,
+             uintN      argc,
+             jsval     *argv,
+             jsval     *retval)
+{
+    char *buffer;
+
+    if (!gjs_print_parse_args(context, argc, argv, &buffer)) {
+        return FALSE;
+    }
+
+    g_printerr("%s\n", buffer);
+    g_free(buffer);
+
+    return JS_TRUE;
+}
+
 static JSClass global_class = {
     "GjsGlobal", JSCLASS_GLOBAL_FLAGS,
     JS_PropertyStub, JS_PropertyStub, JS_PropertyStub, JS_PropertyStub,
@@ -226,6 +324,11 @@ gjs_context_dispose(GObject *object)
 
     js_context = GJS_CONTEXT(object);
 
+    if (js_context->profiler) {
+        gjs_profiler_free(js_context->profiler);
+        js_context->profiler = NULL;
+    }
+
     if (js_context->global != NULL) {
         JS_RemoveRoot(js_context->context, &js_context->global);
         js_context->global = NULL;
@@ -283,10 +386,130 @@ gjs_context_finalize(GObject *object)
         g_strfreev(js_context->search_path);
         js_context->search_path = NULL;
     }
+    
+    g_static_mutex_lock(&contexts_lock);
+    all_contexts = g_list_remove(all_contexts, object);
+    g_static_mutex_unlock(&contexts_lock);
 
     G_OBJECT_CLASS(gjs_context_parent_class)->finalize(object);
 }
 
+/* Implementations of locale-specific operations; these are used
+ * in the implementation of String.localeCompare(), Date.toLocaleDateString(),
+ * and so forth. We take the straight-forward approach of converting
+ * to UTF-8, using the appropriate GLib functions, and converting
+ * back if necessary.
+ */
+static JSBool
+gjs_locale_to_upper_case (JSContext *context,
+                          JSString  *src,
+                          jsval     *retval)
+{
+    JSBool success = JS_FALSE;
+    char *utf8 = NULL;
+    char *upper_case_utf8 = NULL;
+
+    if (!gjs_string_to_utf8(context, STRING_TO_JSVAL(src), &utf8))
+        goto out;
+
+    upper_case_utf8 = g_utf8_strup (utf8, -1);
+
+    if (!gjs_string_from_utf8(context, upper_case_utf8, -1, retval))
+        goto out;
+
+    success = JS_TRUE;
+
+out:
+    g_free(utf8);
+    g_free(upper_case_utf8);
+
+    return success;
+}
+
+static JSBool
+gjs_locale_to_lower_case (JSContext *context,
+                          JSString  *src,
+                          jsval     *retval)
+{
+    JSBool success = JS_FALSE;
+    char *utf8 = NULL;
+    char *lower_case_utf8 = NULL;
+
+    if (!gjs_string_to_utf8(context, STRING_TO_JSVAL(src), &utf8))
+        goto out;
+
+    lower_case_utf8 = g_utf8_strdown (utf8, -1);
+
+    if (!gjs_string_from_utf8(context, lower_case_utf8, -1, retval))
+        goto out;
+
+    success = JS_TRUE;
+
+out:
+    g_free(utf8);
+    g_free(lower_case_utf8);
+
+    return success;
+}
+
+static JSBool
+gjs_locale_compare (JSContext *context,
+                    JSString  *src_1,
+                    JSString  *src_2,
+                    jsval     *retval)
+{
+    JSBool success = JS_FALSE;
+    char *utf8_1 = NULL, *utf8_2 = NULL;
+    int result;
+
+    if (!gjs_string_to_utf8(context, STRING_TO_JSVAL(src_1), &utf8_1) ||
+        !gjs_string_to_utf8(context, STRING_TO_JSVAL(src_2), &utf8_2))
+        goto out;
+
+    result = g_utf8_collate (utf8_1, utf8_2);
+    *retval = INT_TO_JSVAL(result);
+
+    success = JS_TRUE;
+
+out:
+    g_free(utf8_1);
+    g_free(utf8_2);
+
+    return success;
+}
+
+static JSBool
+gjs_locale_to_unicode (JSContext *context,
+                       char      *src,
+                       jsval     *retval)
+{
+    JSBool success;
+    char *utf8;
+    GError *error = NULL;
+
+    utf8 = g_locale_to_utf8(src, -1, NULL, NULL, &error);
+    if (!utf8) {
+        gjs_throw(context,
+                  "Failed to convert locale string to UTF8: %s",
+                  error->message);
+        g_error_free(error);
+        return JS_FALSE;
+    }
+
+    success = gjs_string_from_utf8(context, utf8, -1, retval);
+    g_free (utf8);
+
+    return success;
+}
+
+static JSLocaleCallbacks gjs_locale_callbacks =
+{
+    gjs_locale_to_upper_case,
+    gjs_locale_to_lower_case,
+    gjs_locale_compare,
+    gjs_locale_to_unicode
+};
+
 static GObject*
 gjs_context_constructor (GType                  type,
                          guint                  n_construct_properties,
@@ -294,6 +517,7 @@ gjs_context_constructor (GType                  type,
 {
     GObject *object;
     GjsContext *js_context;
+    guint32 options_flags;
 
     object = (* G_OBJECT_CLASS (gjs_context_parent_class)->constructor) (type,
                                                                          n_construct_properties,
@@ -302,9 +526,10 @@ gjs_context_constructor (GType                  type,
     js_context = GJS_CONTEXT(object);
 
     if (js_context->runtime == NULL) {
-        js_context->runtime = JS_NewRuntime(1024*1024 /* max bytes */);
+        js_context->runtime = JS_NewRuntime(32*1024*1024 /* max bytes */);
         if (js_context->runtime == NULL)
             gjs_fatal("Failed to create javascript runtime");
+        JS_SetGCParameter(js_context->runtime, JSGC_MAX_BYTES, 0xffffffff);
         js_context->we_own_runtime = TRUE;
     }
 
@@ -312,16 +537,29 @@ gjs_context_constructor (GType                  type,
     if (js_context->context == NULL)
         gjs_fatal("Failed to create javascript context");
 
+    /* same as firefox, see discussion at
+     * https://bugzilla.mozilla.org/show_bug.cgi?id=420869 */
+    JS_SetScriptStackQuota(js_context->context, 100*1024*1024);
+
     /* JSOPTION_DONT_REPORT_UNCAUGHT: Don't send exceptions to our
      * error report handler; instead leave them set.  This allows us
      * to get at the exception object.
      *
      * JSOPTION_STRICT: Report warnings to error reporter function.
      */
+    options_flags = JSOPTION_DONT_REPORT_UNCAUGHT | JSOPTION_STRICT;
+
+#ifdef JSOPTION_JIT
+    if (!g_getenv("GJS_DISABLE_JIT")) {
+        gjs_debug(GJS_DEBUG_CONTEXT, "Enabling JIT");
+        options_flags |= JSOPTION_JIT;
+    }
+#endif
+
     JS_SetOptions(js_context->context,
-                  JS_GetOptions(js_context->context) |
-                  JSOPTION_DONT_REPORT_UNCAUGHT |
-                  JSOPTION_STRICT);
+                  JS_GetOptions(js_context->context) | options_flags);
+
+    JS_SetLocaleCallbacks(js_context->context, &gjs_locale_callbacks);
 
     JS_SetErrorReporter(js_context->context, gjs_error_reporter);
 
@@ -373,6 +611,18 @@ gjs_context_constructor (GType                  type,
                            2, GJS_MODULE_PROP_FLAGS))
         gjs_fatal("Failed to define logError function");
 
+    /* Define global functions called print() and printerr() */
+    if (!JS_DefineFunction(js_context->context, js_context->global,
+                           "print",
+                           gjs_print,
+                           3, GJS_MODULE_PROP_FLAGS))
+        gjs_fatal("Failed to define print function");
+    if (!JS_DefineFunction(js_context->context, js_context->global,
+                           "printerr",
+                           gjs_printerr,
+                           4, GJS_MODULE_PROP_FLAGS))
+        gjs_fatal("Failed to define printerr function");
+
     /* If we created the root importer in the load context,
      * there would be infinite recursion since the load context
      * is a GjsContext
@@ -398,6 +648,14 @@ gjs_context_constructor (GType                  type,
             gjs_fatal("Failed to point 'imports' property at root importer");
     }
 
+    if (js_context->we_own_runtime) {
+        js_context->profiler = gjs_profiler_new(js_context->runtime);
+    }
+
+    g_static_mutex_lock (&contexts_lock);
+    all_contexts = g_list_prepend(all_contexts, object);
+    g_static_mutex_unlock (&contexts_lock);
+
     return object;
 }
 
@@ -463,9 +721,38 @@ gjs_context_new_with_search_path(char** search_path)
                          NULL);
 }
 
-JSContext*
-gjs_context_get_context(GjsContext *js_context)
+/**
+ * gjs_context_get_all:
+ * 
+ * Returns a newly-allocated list containing all known instances of #GjsContext.
+ * This is useful for operating on the contexts from a process-global situation
+ * such as a debugger.
+ * 
+ * Return value: (element-type GjsContext) (transfer full): Known #GjsContext instances
+ */
+GList*
+gjs_context_get_all(void)
 {
+  GList *result;
+  GList *iter;
+  g_static_mutex_lock (&contexts_lock);
+  result = g_list_copy(all_contexts);
+  for (iter = result; iter; iter = iter->next)
+    g_object_ref((GObject*)iter->data);
+  g_static_mutex_unlock (&contexts_lock);
+  return result;
+}
+
+/**
+ * gjs_context_get_native_context:
+ *
+ * Returns a pointer to the underlying native context.  For SpiderMonkey, this
+ * is a JSContext *
+ */
+void*
+gjs_context_get_native_context (GjsContext *js_context)
+{
+    g_return_val_if_fail(GJS_IS_CONTEXT(js_context), NULL);
     return js_context->context;
 }
 
@@ -483,6 +770,7 @@ gjs_context_eval(GjsContext *js_context,
                  int          *exit_status_p,
                  GError      **error)
 {
+    int line_number;
     jsval retval;
     gboolean success;
 
@@ -496,6 +784,20 @@ gjs_context_eval(GjsContext *js_context,
      */
     success = TRUE;
 
+    /* handle scripts with UNIX shebangs */
+    line_number = 1;
+    if (script != NULL && script[0] == '#' && script[1] == '!') {
+        const char *s;
+
+        s = (const char *) strstr (script, "\n");
+        if (s != NULL) {
+            if (script_len > 0)
+                script_len -= (s + 1 - script);
+            script = s + 1;
+            line_number = 2;
+        }
+    }
+
     /* log and clear exception if it's set (should not be, normally...) */
     if (gjs_log_exception(js_context->context,
                              NULL)) {
@@ -509,7 +811,7 @@ gjs_context_eval(GjsContext *js_context,
                            script,
                            script_len >= 0 ? script_len : (gssize) strlen(script),
                            filename,
-                           1, /* line number */
+                           line_number,
                            &retval)) {
         char *message;
 
