@@ -1,6 +1,7 @@
 /* -*- mode: C; c-basic-offset: 4; indent-tabs-mode: nil; -*- */
 /*
  * Copyright (c) 2008  litl, LLC
+ * Copyright (c) 2009 Red Hat, Inc.
  *
  * Permission is hereby granted, free of charge, to any person obtaining a copy
  * of this software and associated documentation files (the "Software"), to
@@ -29,10 +30,15 @@
 
 #include "jsapi-util.h"
 #include "context-jsapi.h"
+#include "jsapi-private.h"
 
 #include <string.h>
-#include <jscntxt.h>
 
+GQuark
+gjs_util_error_quark (void)
+{
+  return g_quark_from_static_string ("gjs-util-error-quark");
+}
 
 typedef struct {
     GHashTable *dynamic_classes;
@@ -82,7 +88,7 @@ gjs_runtime_get_load_context(JSRuntime *runtime)
                                 g_object_unref);
     }
 
-    return gjs_context_get_context(context);
+    return (JSContext*)gjs_context_get_native_context(context);
 }
 
 JSContext*
@@ -94,7 +100,7 @@ gjs_runtime_peek_load_context(JSRuntime *runtime)
     if (context == NULL) {
         return NULL;
     } else {
-        return gjs_context_get_context(context);
+        return (JSContext*)gjs_context_get_native_context(context);
     }
 }
 
@@ -138,7 +144,7 @@ gjs_runtime_get_call_context(JSRuntime *runtime)
                                 g_object_unref);
     }
 
-    return gjs_context_get_context(context);
+    return (JSContext*)gjs_context_get_native_context(context);
 }
 
 static JSContext*
@@ -150,7 +156,7 @@ gjs_runtime_peek_call_context(JSRuntime *runtime)
     if (context == NULL) {
         return NULL;
     } else {
-        return gjs_context_get_context(context);
+        return (JSContext*)gjs_context_get_native_context(context);
     }
 }
 
@@ -430,7 +436,7 @@ gjs_get_instance_private_dynamic(JSContext      *context,
         return NULL;
     }
 
-    obj_class = JS_GetClass(context, obj);
+    obj_class = JS_GET_CLASS(context, obj);
     g_assert(obj_class != NULL);
 
     rd = get_data_from_context(context);
@@ -466,7 +472,7 @@ gjs_get_instance_private_dynamic_with_typecheck(JSContext      *context,
         return NULL;
     }
 
-    obj_class = JS_GetClass(context, obj);
+    obj_class = JS_GET_CLASS(context, obj);
     g_assert(obj_class != NULL);
 
     rd = get_data_from_context(context);
@@ -502,7 +508,7 @@ gjs_construct_object_dynamic(JSContext      *context,
      */
     load_context = gjs_runtime_get_load_context(JS_GetRuntime(context));
 
-    proto_class = JS_GetClass(load_context, proto);
+    proto_class = JS_GET_CLASS(load_context, proto);
 
     rd = get_data_from_context(load_context);
 
@@ -592,7 +598,7 @@ gjs_value_debug_string(JSContext      *context,
              */
             JSClass *klass;
 
-            klass = JS_GetClass(context, JSVAL_TO_OBJECT(value));
+            klass = JS_GET_CLASS(context, JSVAL_TO_OBJECT(value));
             if (klass != NULL) {
                 str = JS_NewStringCopyZ(context, klass->name);
                 JS_ClearPendingException(context);
@@ -835,6 +841,44 @@ gjs_log_and_keep_exception(JSContext *context,
     return log_and_maybe_keep_exception(context, message_p, TRUE);
 }
 
+static void
+try_to_chain_stack_trace(JSContext *src_context, JSContext *dst_context,
+                         jsval src_exc) {
+    /* append current stack of dst_context to stack trace for src_exc.
+     * we bail if anything goes wrong, just using the src_exc unmodified
+     * in that case. */
+    jsval chained, src_stack, dst_stack, new_stack;
+    JSString *new_stack_str;
+
+    if (!JSVAL_IS_OBJECT(src_exc))
+        return; // src_exc doesn't have a stack trace
+
+    /* create a new exception in dst_context to get a stack trace */
+    gjs_throw_literal(dst_context, "Chained exception");
+    if (!(JS_GetPendingException(dst_context, &chained) &&
+          JSVAL_IS_OBJECT(chained)))
+        return; // gjs_throw_literal didn't work?!
+    JS_ClearPendingException(dst_context);
+
+    /* get stack trace for src_exc and chained */
+    if (!(gjs_object_get_property(dst_context, JSVAL_TO_OBJECT(chained),
+                                  "stack", &dst_stack) &&
+          JSVAL_IS_STRING(dst_stack)))
+        return; // couldn't get chained stack
+    if (!(gjs_object_get_property(src_context, JSVAL_TO_OBJECT(src_exc),
+                                  "stack", &src_stack) &&
+          JSVAL_IS_STRING(src_stack)))
+        return; // couldn't get source stack
+
+    /* add chained exception's stack trace to src_exc */
+    new_stack_str = JS_ConcatStrings
+        (dst_context, JSVAL_TO_STRING(src_stack), JSVAL_TO_STRING(dst_stack));
+    if (new_stack_str==NULL)
+        return; // couldn't concatenate src and dst stacks?!
+    new_stack = STRING_TO_JSVAL(new_stack_str);
+    JS_SetProperty(dst_context, JSVAL_TO_OBJECT(src_exc), "stack", &new_stack);
+}
+
 JSBool
 gjs_move_exception(JSContext      *src_context,
                    JSContext      *dest_context)
@@ -843,6 +887,10 @@ gjs_move_exception(JSContext      *src_context,
     jsval exc;
     if (JS_GetPendingException(src_context, &exc)) {
         if (src_context != dest_context) {
+            /* try to add the current stack of dest_context to the
+             * stack trace of exc */
+            try_to_chain_stack_trace(src_context, dest_context, exc);
+            /* move the exception to dest_context */
             JS_SetPendingException(dest_context, exc);
             JS_ClearPendingException(src_context);
         }
@@ -872,55 +920,6 @@ gjs_call_function_value(JSContext      *context,
     return result;
 }
 
-void
-gjs_error_reporter(JSContext     *context,
-                   const char    *message,
-                   JSErrorReport *report)
-{
-    const char *warning;
-
-    if (gjs_environment_variable_is_set("GJS_ABORT_ON_OOM") &&
-        report->flags == JSREPORT_ERROR &&
-        report->errorNumber == JSMSG_OUT_OF_MEMORY) {
-        g_error("GJS ran out of memory at %s: %i.",
-                report->filename,
-                report->lineno);
-    }
-
-    if ((report->flags & JSREPORT_WARNING) != 0) {
-        /* We manually insert "WARNING" into the output instead of
-         * having GJS_DEBUG_WARNING because it's convenient to
-         * search for 'JS ERROR' to find all problems
-         */
-        warning = "WARNING: ";
-
-        /* suppress bogus warnings. See mozilla/js/src/js.msg */
-        switch (report->errorNumber) {
-            /* 162, JSMSG_UNDEFINED_PROP: warns every time a lazy property
-             * is resolved, since the property starts out
-             * undefined. When this is a real bug it should usually
-             * fail somewhere else anyhow.
-             */
-        case 162:
-            return;
-        }
-    } else {
-        warning = "REPORTED: ";
-    }
-
-    gjs_debug(GJS_DEBUG_ERROR,
-              "%s'%s'",
-              warning,
-              message);
-
-    gjs_debug(GJS_DEBUG_ERROR,
-              "%sfile '%s' line %u exception %d number %d",
-              warning,
-              report->filename, report->lineno,
-              (report->flags & JSREPORT_EXCEPTION) != 0,
-              report->errorNumber);
-}
-
 static JSBool
 log_prop(JSContext  *context,
          JSObject   *obj,
@@ -1026,7 +1025,7 @@ gjs_date_from_time_t (JSContext *context, time_t time)
     if (!JS_GetProperty(context, date_constructor, "prototype", &date_prototype))
         gjs_fatal("Failed to get prototype from Date constructor");
 
-    date_class = JS_GetClass(context, JSVAL_TO_OBJECT (date_prototype));
+    date_class = JS_GET_CLASS(context, JSVAL_TO_OBJECT (date_prototype));
 
     if (!JS_NewNumberValue(context, ((double) time) * 1000, &(args[0])))
         gjs_fatal("Failed to convert time_t to number");
@@ -1037,3 +1036,196 @@ gjs_date_from_time_t (JSContext *context, time_t time)
     JS_LeaveLocalRootScope(context);
     return OBJECT_TO_JSVAL(date);
 }
+
+/**
+ * gjs_parse_args:
+ * @context:
+ * @function_name: The name of the function being called
+ * @format: Printf-like format specifier containing the expected arguments
+ * @argc: Number of JavaScript arguments
+ * @argv: JavaScript argument array
+ * @Varargs: for each character in @format, a pair of a char * which is the name
+ * of the argument, and a pointer to a location to store the value. The type of
+ * value stored depends on the format character, as described below.
+ * 
+ * This function is inspired by Python's PyArg_ParseTuple for those
+ * familiar with it.  It takes a format specifier which gives the
+ * types of the expected arguments, and a list of argument names and
+ * value location pairs.  The currently accepted format specifiers are:
+ * 
+ * s: A string, converted into UTF-8
+ * z: Like 's', but may be null in JavaScript (which appears as NULL in C)
+ * F: A string, converted into "filename encoding" (i.e. active locale)
+ * i: A number, will be converted to a C "gint32"
+ * u: A number, converted into a C "guint32"
+ * o: A JavaScript object, as a "JSObject *"
+ *
+ * The '|' character introduces optional arguments.  All format specifiers
+ * after a '|' when not specified, do not cause any changes in the C
+ * value location.
+ */
+JSBool
+gjs_parse_args (JSContext  *context,
+                const char *function_name,
+                const char *format,
+                uintN      argc,
+                jsval     *argv,
+                ...)
+{
+    guint i;
+    const char *fmt_iter;
+    va_list args;
+    GError *arg_error = NULL;
+    guint n_unwind = 0;
+#define MAX_UNWIND_STRINGS 16
+    gpointer unwind_strings[MAX_UNWIND_STRINGS];
+    guint n_required;
+    guint n_total;
+    guint consumed_args;
+
+    va_start (args, argv);
+
+    /* Check for optional argument specifier */
+    fmt_iter = strchr (format, '|');
+    if (fmt_iter) {
+        /* Be sure there's not another '|' */
+        g_return_val_if_fail (strchr (fmt_iter + 1, '|') == NULL, JS_FALSE);
+
+        n_required = fmt_iter - format;
+        n_total = n_required + strlen (fmt_iter + 1);
+    } else {
+        n_required = n_total = strlen (format);
+    }
+
+    if (argc < n_required || argc > n_total) {
+        if (n_required == n_total) {
+            gjs_throw(context, "Error invoking %s: Expected %d arguments, got %d", function_name,
+                      n_required, argc);
+        } else {
+            gjs_throw(context, "Error invoking %s: Expected minimum %d arguments (and %d optional), got %d", function_name,
+                      n_required, n_total - n_required, argc);
+        }
+        goto error_unwind;
+    }
+
+    /* We have 3 iteration variables here.
+     * @i: The current integer position in fmt_args
+     * @fmt_iter: A pointer to the character in fmt_args
+     * @consumed_args: How many arguments we've taken from argv
+     * 
+     * consumed_args can currently be different from 'i' because of the '|' character.
+     */
+    for (i = 0, consumed_args = 0, fmt_iter = format; *fmt_iter; fmt_iter++, i++) {
+        const char *argname;
+        gpointer arg_location;
+        jsval js_value;
+        const char *arg_error_message = NULL;
+
+        if (*fmt_iter == '|')
+            continue;
+
+        if (consumed_args == argc) {
+            break;
+        }
+
+        argname = va_arg (args, char *);
+        arg_location = va_arg (args, gpointer);
+
+        g_return_val_if_fail (argname != NULL, JS_FALSE);
+        g_return_val_if_fail (arg_location != NULL, JS_FALSE);
+
+        js_value = argv[consumed_args];
+        
+        switch (*fmt_iter) {
+            case 'o': {
+               if (!JSVAL_IS_OBJECT(js_value)) {
+                   arg_error_message = "Not an object";
+               } else {
+                   JSObject **arg = arg_location;
+                   *arg = JSVAL_TO_OBJECT(js_value);
+               }
+            }
+            break;
+            case 's':
+            case 'z': {
+                char **arg = arg_location;
+                
+                if (*fmt_iter == 'z' && JSVAL_IS_NULL(js_value)) {
+                    *arg = NULL;
+                } else {
+                    if (gjs_try_string_to_utf8 (context, js_value, arg, &arg_error)) {
+                        unwind_strings[n_unwind++] = *arg;
+                        g_assert(n_unwind < MAX_UNWIND_STRINGS);
+                    }
+                }    
+            }
+            break;
+            case 'F': {
+                char **arg = arg_location;
+                
+                if (gjs_try_string_to_filename (context, js_value, arg, &arg_error)) {
+                    unwind_strings[n_unwind++] = *arg;
+                    g_assert(n_unwind < MAX_UNWIND_STRINGS);
+                }
+            }
+            break;
+            case 'i': {
+                if (!JS_ValueToInt32(context, js_value, (gint32*) arg_location)) {
+                    /* Our error message is going to be more useful */
+                    JS_ClearPendingException(context);
+                    arg_error_message = "Couldn't convert to integer";
+                }
+            }
+            break;
+            case 'u': {
+                gdouble num;
+                if (!JS_ValueToNumber(context, js_value, &num)) {
+                    /* Our error message is going to be more useful */
+                    JS_ClearPendingException(context);
+                    arg_error_message = "Couldn't convert to unsigned integer";
+                } else if (num > G_MAXUINT32 || num < 0) {
+                    arg_error_message = "Value is out of range";
+                } else {
+                    *((guint32*) arg_location) = num;
+                }
+            }
+            break;
+            case 'f': {
+                double num;
+                if (!JS_ValueToNumber(context, js_value, &num)) {
+                    /* Our error message is going to be more useful */
+                    JS_ClearPendingException(context);
+                    arg_error_message = "Couldn't convert to double";
+                } else {
+                    *((double*) arg_location) = num;
+                }
+            }
+            break;
+            default:
+              g_assert_not_reached ();
+        }
+        
+        if (arg_error != NULL)
+            arg_error_message = arg_error->message;
+        
+        if (arg_error_message != NULL) {
+            gjs_throw(context, "Error invoking %s, at argument %d (%s): %s", function_name, 
+                      consumed_args+1, argname, arg_error_message);
+            g_clear_error (&arg_error);
+            goto error_unwind;
+        }
+        
+        consumed_args++;
+    }
+
+    va_end (args);
+    return JS_TRUE;
+
+  error_unwind:
+    va_end (args);
+    /* We still own the strings in the error case, free any we converted */
+    for (i = 0; i < n_unwind; i++) {
+        g_free (unwind_strings[i]);
+    }
+    return JS_FALSE;
+}
